@@ -66,7 +66,7 @@
 #define IO_CAPABILITY_DISPLAYONLY	0x00
 #define IO_CAPABILITY_DISPLAYYESNO	0x01
 #define IO_CAPABILITY_KEYBOARDONLY	0x02
-#define IO_CAPABILITY_NOINPUTOUTPUT	0x03
+#define IO_CAPABILITY_NOINPUTNOOUTPUT	0x03
 #define IO_CAPABILITY_INVALID		0xFF
 
 #define check_address(address) bachk(address)
@@ -128,6 +128,7 @@ struct btd_adapter {
 	gboolean off_requested;		/* DEVDOWN ioctl was called */
 
 	uint8_t svc_cache;		/* For bluetoothd startup */
+	gint ref;
 };
 
 static void adapter_set_pairable_timeout(struct btd_adapter *adapter,
@@ -213,82 +214,140 @@ static void dev_info_free(struct remote_dev_info *dev)
 	g_free(dev);
 }
 
+void clear_found_devices_list(struct btd_adapter *adapter)
+{
+	if (!adapter->found_devices)
+		return;
+
+	g_slist_foreach(adapter->found_devices, (GFunc) dev_info_free, NULL);
+	g_slist_free(adapter->found_devices);
+	adapter->found_devices = NULL;
+}
+
+static int set_service_classes(struct btd_adapter *adapter, uint8_t value)
+{
+	struct hci_dev *dev = &adapter->dev;
+	const uint8_t *cls = dev->class;
+	uint32_t dev_class;
+	int dd, err;
+
+	if (cls[2] == value)
+		return 0; /* Already set */
+
+	dd = hci_open_dev(adapter->dev_id);
+	if (dd < 0) {
+		err = -errno;
+		error("Can't open device hci%d: %s (%d)",
+				adapter->dev_id, strerror(errno), errno);
+		return err;
+	}
+
+	dev_class = (value << 16) | (cls[1] << 8) | cls[0];
+
+	debug("Changing service classes to 0x%06x", dev_class);
+
+	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
+		err = -errno;
+		error("Can't write class of device: %s (%d)",
+						strerror(errno), errno);
+		hci_close_dev(dd);
+		return err;
+	}
+
+	hci_close_dev(dd);
+
+	return 0;
+}
+
+int set_major_and_minor_class(struct btd_adapter *adapter, uint8_t major,
+								uint8_t minor)
+{
+	struct hci_dev *dev = &adapter->dev;
+	const uint8_t *cls = dev->class;
+	uint32_t dev_class;
+	int dd, err;
+
+	dd = hci_open_dev(adapter->dev_id);
+	if (dd < 0) {
+		err = -errno;
+		error("Can't open device hci%d: %s (%d)",
+				adapter->dev_id, strerror(errno), errno);
+		return err;
+	}
+
+	dev_class = (cls[2] << 16) | ((cls[1] & 0x20) << 8) |
+						((major & 0xdf) << 8) | minor;
+
+	debug("Changing major/minor class to 0x%06x", dev_class);
+
+	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
+		int err = -errno;
+		error("Can't write class of device: %s (%d)",
+						strerror(errno), errno);
+		hci_close_dev(dd);
+		return err;
+	}
+
+	hci_close_dev(dd);
+	return 0;
+}
+
 int pending_remote_name_cancel(struct btd_adapter *adapter)
 {
 	struct remote_dev_info *dev, match;
-	GSList *l;
-	int dd, err = 0;
+	int err = 0;
 
 	/* find the pending remote name request */
 	memset(&match, 0, sizeof(struct remote_dev_info));
 	bacpy(&match.bdaddr, BDADDR_ANY);
 	match.name_status = NAME_REQUESTED;
 
-	l = g_slist_find_custom(adapter->found_devices, &match,
-			(GCompareFunc) found_device_cmp);
-	if (!l) /* no pending request */
-		return 0;
+	dev = adapter_search_found_devices(adapter, &match);
+	if (!dev) /* no pending request */
+		return -ENODATA;
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0)
-		return -ENODEV;
-
-	dev = l->data;
-
-	if (hci_read_remote_name_cancel(dd, &dev->bdaddr,
-					HCI_REQ_TIMEOUT) < 0) {
-		err = -errno;
+	err = adapter_ops->cancel_resolve_name(adapter->dev_id, &dev->bdaddr);
+	if (err < 0)
 		error("Remote name cancel failed: %s(%d)",
 						strerror(errno), errno);
-	}
-
-	/* free discovered devices list */
-	g_slist_foreach(adapter->found_devices, (GFunc) dev_info_free, NULL);
-	g_slist_free(adapter->found_devices);
-	adapter->found_devices = NULL;
-
-	hci_close_dev(dd);
-
 	return err;
 }
 
-static int set_limited_discoverable(int dd, const uint8_t *cls,
-							gboolean limited)
+int adapter_resolve_names(struct btd_adapter *adapter)
 {
-	uint32_t dev_class;
-	int num = (limited ? 2 : 1);
-	uint8_t lap[] = { 0x33, 0x8b, 0x9e, 0x00, 0x8b, 0x9e };
-	/*
-	 * 1: giac
-	 * 2: giac + liac
-	 */
-	if (hci_write_current_iac_lap(dd, num, lap, HCI_REQ_TIMEOUT) < 0) {
-		int err = -errno;
-		error("Can't write current IAC LAP: %s(%d)",
+	struct remote_dev_info *dev, match;
+	int err;
+
+	memset(&match, 0, sizeof(struct remote_dev_info));
+	bacpy(&match.bdaddr, BDADDR_ANY);
+	match.name_status = NAME_REQUIRED;
+
+	dev = adapter_search_found_devices(adapter, &match);
+	if (!dev)
+		return -ENODATA;
+
+	/* send at least one request or return failed if the list is empty */
+	do {
+		/* flag to indicate the current remote name requested */
+		dev->name_status = NAME_REQUESTED;
+
+		err = adapter_ops->resolve_name(adapter->dev_id, &dev->bdaddr);
+
+		if (!err)
+			break;
+
+		error("Unable to send HCI remote name req: %s (%d)",
 						strerror(errno), errno);
-		return err;
-	}
 
-	if (limited) {
-		if (cls[1] & 0x20)
-			return 0; /* Already limited */
+		/* if failed, request the next element */
+		/* remove the element from the list */
+		adapter_remove_found_device(adapter, &dev->bdaddr);
 
-		dev_class = (cls[2] << 16) | ((cls[1] | 0x20) << 8) | cls[0];
-	} else {
-		if (!(cls[1] & 0x20))
-			return 0; /* Already clear */
+		/* get the next element */
+		dev = adapter_search_found_devices(adapter, &match);
+	} while (dev);
 
-		dev_class = (cls[2] << 16) | ((cls[1] & 0xdf) << 8) | cls[0];
-	}
-
-	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
-		int err = -errno;
-		error("Can't write class of device: %s (%d)",
-						strerror(errno), errno);
-		return err;
-	}
-
-	return 0;
+	return err;
 }
 
 static const char *mode2str(uint8_t mode)
@@ -343,23 +402,10 @@ static void adapter_remove_discov_timeout(struct btd_adapter *adapter)
 static gboolean discov_timeout_handler(gpointer user_data)
 {
 	struct btd_adapter *adapter = user_data;
-	int dd;
-	uint8_t scan_enable;
-	uint16_t dev_id = adapter->dev_id;
 
 	adapter->discov_timeout_id = 0;
 
-	dd = hci_open_dev(dev_id);
-	if (dd < 0) {
-		error("HCI device open failed: hci%d", dev_id);
-		return FALSE;
-	}
-
-	scan_enable = adapter->scan_mode & ~SCAN_INQUIRY;
-
-	hci_send_cmd(dd, OGF_HOST_CTL, OCF_WRITE_SCAN_ENABLE,
-					1, &scan_enable);
-	hci_close_dev(dd);
+	adapter_ops->set_connectable(adapter->dev_id);
 
 	return FALSE;
 }
@@ -380,89 +426,58 @@ static void adapter_set_discov_timeout(struct btd_adapter *adapter,
 							adapter);
 }
 
-static uint8_t mode2scan(uint8_t mode)
-{
-	switch (mode) {
-	case MODE_OFF:
-		return SCAN_DISABLED;
-	case MODE_CONNECTABLE:
-		return SCAN_PAGE;
-	case MODE_DISCOVERABLE:
-	case MODE_LIMITED:
-		return (SCAN_PAGE | SCAN_INQUIRY);
-	default:
-		error("Invalid mode given to mode2scan: %u", mode);
-		return SCAN_PAGE;
-	}
-}
-
 static int set_mode(struct btd_adapter *adapter, uint8_t new_mode)
 {
-	uint8_t scan_enable;
-	uint8_t current_scan = adapter->scan_mode;
-	int err, dd;
+	int err;
 	const char *modestr;
 
-	scan_enable = mode2scan(new_mode);
-
-	if (!adapter->up && scan_enable != SCAN_DISABLED) {
-		err = adapter_ops->start(adapter->dev_id);
+	if (!adapter->up && new_mode != MODE_OFF) {
+		err = adapter_ops->set_powered(adapter->dev_id, TRUE);
 		if (err < 0)
 			return err;
 	}
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0)
-		return -EIO;
-
-	if (adapter->up && scan_enable == SCAN_DISABLED) {
-		struct hci_request rq = {
-			.ogf = OGF_HOST_CTL,
-			.ocf = OCF_WRITE_SCAN_ENABLE,
-			.cparam = &scan_enable,
-			.clen = sizeof(scan_enable),
-		};
-
-		hci_send_req(dd, &rq, HCI_REQ_TIMEOUT);
-
-		if (ioctl(dd, HCIDEVDOWN, adapter->dev_id) < 0) {
-			err = -errno;
-			hci_close_dev(dd);
+	if (adapter->up && new_mode == MODE_OFF) {
+		err = adapter_ops->set_powered(adapter->dev_id, FALSE);
+		if (err < 0)
 			return err;
-		}
 
 		adapter->off_requested = TRUE;
 
 		goto done;
 	}
 
-	if (current_scan != scan_enable) {
-		err = hci_send_cmd(dd, OGF_HOST_CTL, OCF_WRITE_SCAN_ENABLE,
-					1, &scan_enable);
-		if (err < 0) {
-			hci_close_dev(dd);
-			return err;
-		}
-	} else if ((scan_enable & SCAN_INQUIRY) &&
-						(new_mode != adapter->mode)) {
+	if (new_mode == adapter->mode)
+		return 0;
+
+	if (new_mode == MODE_CONNECTABLE)
+		err = adapter_ops->set_connectable(adapter->dev_id);
+	else
+		err = adapter_ops->set_discoverable(adapter->dev_id);
+
+	if (err < 0)
+		return err;
+
+	if (new_mode > MODE_CONNECTABLE) {
 		adapter_remove_discov_timeout(adapter);
+
 		if (adapter->discov_timeout)
 			adapter_set_discov_timeout(adapter,
 						adapter->discov_timeout);
+
 		if (new_mode == MODE_LIMITED)
-			set_limited_discoverable(dd, adapter->dev.class,
-									TRUE);
+			adapter_ops->set_limited_discoverable(adapter->dev_id,
+						adapter->dev.class, TRUE);
 		else if (adapter->mode == MODE_LIMITED)
-			set_limited_discoverable(dd, adapter->dev.class,
-									FALSE);
+			adapter_ops->set_limited_discoverable(adapter->dev_id,
+						adapter->dev.class,FALSE);
 	}
+
 done:
 	modestr = mode2str(new_mode);
 
 	write_device_mode(&adapter->bdaddr, modestr);
 
-	hci_close_dev(dd);
-
 	adapter->mode = new_mode;
 
 	return 0;
@@ -628,17 +643,17 @@ static void session_remove(struct session_req *req)
 
 		debug("Stopping discovery");
 
-		g_slist_foreach(adapter->found_devices, (GFunc) dev_info_free,
-				NULL);
-		g_slist_free(adapter->found_devices);
-		adapter->found_devices = NULL;
+		pending_remote_name_cancel(adapter);
 
-		if (adapter->state & STD_INQUIRY)
-			cancel_discovery(adapter);
-		else if (adapter->scheduler_id)
+		clear_found_devices_list(adapter);
+
+		g_slist_free(adapter->oor_devices);
+		adapter->oor_devices = NULL;
+
+		if (adapter->scheduler_id)
 			g_source_remove(adapter->scheduler_id);
-		else
-			cancel_periodic_discovery(adapter);
+
+		adapter_ops->stop_discovery(adapter->dev_id);
 	}
 }
 
@@ -797,15 +812,21 @@ static DBusMessage *set_pairable_timeout(DBusConnection *conn,
 	return dbus_message_new_method_return(msg);
 }
 
-static void update_ext_inquiry_response(int dd, struct hci_dev *dev)
+static void update_ext_inquiry_response(struct btd_adapter *adapter)
 {
 	uint8_t fec = 0, data[240];
+	struct hci_dev *dev = &adapter->dev;
+	int dd;
 
 	if (!(dev->features[6] & LMP_EXT_INQ))
 		return;
 
 	memset(data, 0, sizeof(data));
 
+	dd = hci_open_dev(adapter->dev_id);
+	if (dd < 0)
+		return;
+
 	if (dev->ssp_mode > 0)
 		create_ext_inquiry_response((char *) dev->name, data);
 
@@ -813,83 +834,90 @@ static void update_ext_inquiry_response(int dd, struct hci_dev *dev)
 						HCI_REQ_TIMEOUT) < 0)
 		error("Can't write extended inquiry response: %s (%d)",
 						strerror(errno), errno);
-}
-
-void adapter_name_changed(struct btd_adapter *adapter, const char *name)
-{
-	struct hci_dev *dev = &adapter->dev;
-	int dd;
-
-	if (strncmp(name, (char *) dev->name, 248) == 0)
-		return;
-
-	write_local_name(&adapter->bdaddr, (char *) name);
-
-	strncpy((char *) dev->name, name, 248);
-
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd >= 0) {
-		update_ext_inquiry_response(dd, dev);
-		hci_close_dev(dd);
-	}
-
-	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
-				"Name", DBUS_TYPE_STRING, &name);
-}
-
-static int adapter_set_name(struct btd_adapter *adapter, const char *name)
-{
-	struct hci_dev *dev = &adapter->dev;
-	int dd, err;
-
-	write_local_name(&adapter->bdaddr, (char *) name);
-
-	strncpy((char *) dev->name, name, 248);
-
-	if (!adapter->up)
-		return 0;
-
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0) {
-		err = -errno;
-		error("Can't open device hci%d: %s (%d)",
-				adapter->dev_id, strerror(errno), errno);
-		return err;
-	}
-
-	if (hci_write_local_name(dd, name, HCI_REQ_TIMEOUT) < 0) {
-		err = -errno;
-		error("Can't write name for hci%d: %s (%d)",
-				adapter->dev_id, strerror(errno), errno);
-		hci_close_dev(dd);
-		return err;
-	}
-
-	update_ext_inquiry_response(dd, dev);
 
 	hci_close_dev(dd);
+}
+
+void adapter_update_local_name(bdaddr_t *bdaddr, uint8_t status, void *ptr)
+{
+	read_local_name_rp rp;
+	struct hci_dev *dev;
+	struct btd_adapter *adapter;
+	gchar *name;
+
+	if (status)
+		return;
+
+	adapter = manager_find_adapter(bdaddr);
+	if (!adapter) {
+		error("Unable to find matching adapter");
+		return;
+	}
+
+	dev = &adapter->dev;
+
+	memcpy(&rp, ptr, MAX_NAME_LENGTH);
+	if (strncmp((char *) rp.name, (char *) dev->name, MAX_NAME_LENGTH) == 0)
+		return;
+
+	strncpy((char *) dev->name, (char *) rp.name, MAX_NAME_LENGTH);
+
+	write_local_name(bdaddr, (char *) dev->name);
+
+	update_ext_inquiry_response(adapter);
+
+	name = g_strdup((char *) dev->name);
+
+	if (connection)
+		emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
+				"Name", DBUS_TYPE_STRING, &name);
+	g_free(name);
+}
+
+void adapter_setname_complete(bdaddr_t *local, uint8_t status)
+{
+	struct btd_adapter *adapter;
+	int err;
+
+	if (status)
+		return;
+
+	adapter = manager_find_adapter(local);
+	if (!adapter) {
+		error("No matching adapter found");
+		return;
+	}
+
+	err = adapter_ops->read_name(adapter->dev_id);
+	if (err < 0)
+		error("Sending getting name command failed: %s (%d)",
+						strerror(errno), errno);
 
-	return 0;
 }
 
 static DBusMessage *set_name(DBusConnection *conn, DBusMessage *msg,
 					const char *name, void *data)
 {
 	struct btd_adapter *adapter = data;
-	int ecode;
+	struct hci_dev *dev = &adapter->dev;
+	int err;
 
 	if (!g_utf8_validate(name, -1, NULL)) {
 		error("Name change failed: supplied name isn't valid UTF-8");
 		return invalid_args(msg);
 	}
 
-	ecode = adapter_set_name(adapter, name);
-	if (ecode < 0)
-		return failed_strerror(msg, -ecode);
+	if (strncmp(name, (char *) dev->name, MAX_NAME_LENGTH) == 0)
+		goto done;
 
-	emit_property_changed(conn, adapter->path, ADAPTER_INTERFACE,
-				"Name", DBUS_TYPE_STRING, &name);
+	if (!adapter->up)
+		return failed_strerror(msg, -EHOSTDOWN);
 
+	err = adapter_ops->set_name(adapter->dev_id, name);
+	if (err < 0)
+		return failed_strerror(msg, err);
+
+done:
 	return dbus_message_new_method_return(msg);
 }
 
@@ -1025,104 +1053,16 @@ struct btd_device *adapter_get_device(DBusConnection *conn,
 	return adapter_create_device(conn, adapter, address);
 }
 
-static int start_inquiry(struct btd_adapter *adapter)
+static int adapter_start_inquiry(struct btd_adapter *adapter)
 {
-	inquiry_cp cp;
-	evt_cmd_status rp;
-	struct hci_request rq;
-	uint8_t lap[3] = { 0x33, 0x8b, 0x9e };
-	int dd, err;
+	gboolean periodic = TRUE;
+
+	if (main_opts.discov_interval)
+		periodic = FALSE;
 
 	pending_remote_name_cancel(adapter);
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0)
-		return dd;
-
-	memset(&cp, 0, sizeof(cp));
-	memcpy(&cp.lap, lap, 3);
-	cp.length = 0x08;
-	cp.num_rsp = 0x00;
-
-	memset(&rq, 0, sizeof(rq));
-	rq.ogf = OGF_LINK_CTL;
-	rq.ocf = OCF_INQUIRY;
-	rq.cparam = &cp;
-	rq.clen = INQUIRY_CP_SIZE;
-	rq.rparam = &rp;
-	rq.rlen = EVT_CMD_STATUS_SIZE;
-	rq.event = EVT_CMD_STATUS;
-
-	if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
-		err = -errno;
-		error("Unable to start inquiry: %s (%d)",
-						strerror(errno), errno);
-		hci_close_dev(dd);
-		return err;
-	}
-
-	if (rp.status) {
-		error("HCI_Inquiry command failed with status 0x%02x",
-			rp.status);
-		hci_close_dev(dd);
-		return -bt_error(rp.status);
-	}
-
-	hci_close_dev(dd);
-
-	adapter->state |= RESOLVE_NAME;
-
-	return 0;
-}
-
-static int start_periodic_inquiry(struct btd_adapter *adapter)
-{
-	periodic_inquiry_cp cp;
-	struct hci_request rq;
-	uint8_t lap[3] = { 0x33, 0x8b, 0x9e };
-	uint8_t status;
-	int dd, err;
-
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0)
-		return dd;
-
-	memset(&cp, 0, sizeof(cp));
-	memcpy(&cp.lap, lap, 3);
-	cp.max_period = htobs(24);
-	cp.min_period = htobs(16);
-	cp.length  = 0x08;
-	cp.num_rsp = 0x00;
-
-	memset(&rq, 0, sizeof(rq));
-	rq.ogf    = OGF_LINK_CTL;
-	rq.ocf    = OCF_PERIODIC_INQUIRY;
-	rq.cparam = &cp;
-	rq.clen   = PERIODIC_INQUIRY_CP_SIZE;
-	rq.rparam = &status;
-	rq.rlen   = sizeof(status);
-	rq.event  = EVT_CMD_COMPLETE;
-
-	if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
-		err = -errno;
-		error("Unable to start periodic inquiry: %s (%d)",
-						strerror(errno), errno);
-		hci_close_dev(dd);
-		return err;
-	}
-
-	if (status) {
-		error("HCI_Periodic_Inquiry_Mode failed with status 0x%02x",
-				status);
-		hci_close_dev(dd);
-		return -bt_error(status);
-	}
-
-	hci_close_dev(dd);
-
-	adapter->state |= RESOLVE_NAME;
-
-	return 0;
+	return adapter_ops->start_discovery(adapter->dev_id, periodic);
 }
 
 static DBusMessage *adapter_start_discovery(DBusConnection *conn,
@@ -1145,11 +1085,10 @@ static DBusMessage *adapter_start_discovery(DBusConnection *conn,
 	if (adapter->disc_sessions)
 		goto done;
 
-	if (main_opts.inqmode)
-		err = start_inquiry(adapter);
-	else
-		err = start_periodic_inquiry(adapter);
+	if (main_opts.name_resolv)
+		adapter->state |= RESOLVE_NAME;
 
+	err = adapter_start_inquiry(adapter);
 	if (err < 0)
 		return failed_strerror(msg, -err);
 
@@ -1178,7 +1117,7 @@ static DBusMessage *adapter_stop_discovery(DBusConnection *conn,
 				"Invalid discovery session");
 
 	session_unref(req);
-
+	info("Stopping discovery");
 	return dbus_message_new_method_return(msg);
 }
 
@@ -1195,7 +1134,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	DBusMessage *reply;
 	DBusMessageIter iter;
 	DBusMessageIter dict;
-	char str[249], srcaddr[18];
+	char str[MAX_NAME_LENGTH + 1], srcaddr[18];
 	uint32_t class;
 	gboolean value;
 	char **devices;
@@ -1224,7 +1163,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 
 	/* Name */
 	memset(str, 0, sizeof(str));
-	strncpy(str, (char *) adapter->dev.name, 248);
+	strncpy(str, (char *) adapter->dev.name, MAX_NAME_LENGTH);
 	property = str;
 
 	dict_append_entry(&dict, "Name", DBUS_TYPE_STRING, &property);
@@ -1471,14 +1410,21 @@ static DBusMessage *cancel_device_creation(DBusConnection *conn,
 		return invalid_args(msg);
 
 	device = adapter_find_device(adapter, address);
-	if (!device || !device_is_temporary(device))
+	if (!device || !device_is_creating(device, NULL))
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".NotInProgress",
 				"Device creation not in progress");
 
-	if (!device_is_bonding(device, sender))
+	if (!device_is_creating(device, sender))
 		return not_authorized(msg);
 
+	device_set_temporary(device, TRUE);
+
+	if (device_is_connected(device)) {
+		device_request_disconnect(device, msg);
+		return NULL;
+	}
+
 	adapter_remove_device(conn, adapter, device);
 
 	return dbus_message_new_method_return(msg);
@@ -1509,8 +1455,6 @@ static DBusMessage *create_device(DBusConnection *conn,
 	if (!device)
 		return NULL;
 
-	device_set_temporary(device, FALSE);
-
 	device_browse(device, conn, msg, NULL, FALSE);
 
 	return NULL;
@@ -1526,8 +1470,8 @@ static uint8_t parse_io_capability(const char *capability)
 		return IO_CAPABILITY_DISPLAYYESNO;
 	if (g_str_equal(capability, "KeyboardOnly"))
 		return IO_CAPABILITY_KEYBOARDONLY;
-	if (g_str_equal(capability, "NoInputOutput"))
-		return IO_CAPABILITY_NOINPUTOUTPUT;
+	if (g_str_equal(capability, "NoInputNoOutput"))
+		return IO_CAPABILITY_NOINPUTNOOUTPUT;
 	return IO_CAPABILITY_INVALID;
 }
 
@@ -1560,6 +1504,10 @@ static DBusMessage *create_paired_device(DBusConnection *conn,
 		return invalid_args(msg);
 
 	device = adapter_get_device(conn, adapter, address);
+	if (!device)
+		return g_dbus_create_error(msg,
+				ERROR_INTERFACE ".Failed",
+				"Unable to create a new device object");
 
 	return device_create_bonding(device, conn, msg, agent_path, cap);
 }
@@ -1598,12 +1546,13 @@ static DBusMessage *remove_device(DBusConnection *conn,
 
 	device_set_temporary(device, TRUE);
 
-	if (device_is_connected(device))
-		device_disconnect(device);
-	else
+	if (!device_is_connected(device)) {
 		adapter_remove_device(conn, adapter, device);
+		return dbus_message_new_method_return(msg);
+	}
 
-	return dbus_message_new_method_return(msg);
+	device_request_disconnect(device, msg);
+	return NULL;
 }
 
 static DBusMessage *find_device(DBusConnection *conn,
@@ -1724,8 +1673,10 @@ static GDBusMethodTable adapter_methods[] = {
 						G_DBUS_METHOD_FLAG_ASYNC},
 	{ "CreatePairedDevice",	"sos",	"o",	create_paired_device,
 						G_DBUS_METHOD_FLAG_ASYNC},
-	{ "CancelDeviceCreation","s",	"",	cancel_device_creation	},
-	{ "RemoveDevice",	"o",	"",	remove_device		},
+	{ "CancelDeviceCreation","s",	"",	cancel_device_creation,
+						G_DBUS_METHOD_FLAG_ASYNC},
+	{ "RemoveDevice",	"o",	"",	remove_device,
+						G_DBUS_METHOD_FLAG_ASYNC},
 	{ "FindDevice",		"s",	"o",	find_device		},
 	{ "RegisterAgent",	"os",	"",	register_agent		},
 	{ "UnregisterAgent",	"o",	"",	unregister_agent	},
@@ -1794,15 +1745,23 @@ static int adapter_read_bdaddr(uint16_t dev_id, bdaddr_t *bdaddr)
 	return 0;
 }
 
-static int adapter_setup(struct btd_adapter *adapter, int dd)
+static int adapter_setup(struct btd_adapter *adapter)
 {
 	struct hci_dev *dev = &adapter->dev;
 	uint8_t events[8] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0x00, 0x00 };
 	uint8_t inqmode;
-	int err;
-	char name[249];
+	int err , dd;
+	char name[MAX_NAME_LENGTH + 1];
 
-	if (dev->hci_rev > 1) {
+	dd = hci_open_dev(adapter->dev_id);
+	if (dd < 0) {
+		err = -errno;
+		error("Can't open device hci%d: %s (%d)", adapter->dev_id,
+						strerror(errno), errno);
+		return err;
+	}
+
+	if (dev->lmp_ver > 1) {
 		if (dev->features[5] & LMP_SNIFF_SUBR)
 			events[5] |= 0x20;
 
@@ -1835,16 +1794,12 @@ static int adapter_setup(struct btd_adapter *adapter, int dd)
 						sizeof(events), events);
 	}
 
-	if (read_local_name(&adapter->bdaddr, name) == 0) {
-		memcpy(dev->name, name, 248);
-		hci_write_local_name(dd, name, HCI_REQ_TIMEOUT);
-	}
-
-	update_ext_inquiry_response(dd, dev);
+	if (read_local_name(&adapter->bdaddr, name) == 0)
+		adapter_ops->set_name(adapter->dev_id, name);
 
 	inqmode = get_inquiry_mode(dev);
 	if (inqmode < 1)
-		return 0;
+		goto done;
 
 	if (hci_write_inquiry_mode(dd, inqmode, HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
@@ -1854,6 +1809,8 @@ static int adapter_setup(struct btd_adapter *adapter, int dd)
 		return err;
 	}
 
+done:
+	hci_close_dev(dd);
 	return 0;
 }
 
@@ -1949,11 +1906,15 @@ static void load_drivers(struct btd_adapter *adapter)
 	}
 }
 
-static void load_connections(struct btd_adapter *adapter, int dd)
+static void load_connections(struct btd_adapter *adapter)
 {
 	struct hci_conn_list_req *cl = NULL;
 	struct hci_conn_info *ci;
-	int i;
+	int i, dd;
+
+	dd = hci_open_dev(adapter->dev_id);
+	if (dd < 0)
+		return;
 
 	cl = g_malloc0(10 * sizeof(*ci) + sizeof(*cl));
 
@@ -1963,6 +1924,7 @@ static void load_connections(struct btd_adapter *adapter, int dd)
 
 	if (ioctl(dd, HCIGETCONNLIST, cl) != 0) {
 		g_free(cl);
+		hci_close_dev(dd);
 		return;
 	}
 
@@ -1977,6 +1939,7 @@ static void load_connections(struct btd_adapter *adapter, int dd)
 	}
 
 	g_free(cl);
+	hci_close_dev(dd);
 }
 
 static int get_discoverable_timeout(const char *src)
@@ -1999,11 +1962,12 @@ static int get_pairable_timeout(const char *src)
 	return main_opts.pairto;
 }
 
-static int adapter_up(struct btd_adapter *adapter, int dd)
+static int adapter_up(struct btd_adapter *adapter)
 {
 	char mode[14], srcaddr[18];
 	uint8_t scan_mode;
 	gboolean powered, dev_down = FALSE;
+	int err;
 
 	ba2str(&adapter->bdaddr, srcaddr);
 
@@ -2046,7 +2010,7 @@ static int adapter_up(struct btd_adapter *adapter, int dd)
 		else
 			write_device_mode(&adapter->bdaddr, mode);
 
-		return adapter_up(adapter, dd);
+		return adapter_up(adapter);
 	} else if (!g_str_equal(mode, "connectable") &&
 			adapter->discov_timeout == 0) {
 		/* Set discoverable only if timeout is 0 */
@@ -2056,9 +2020,13 @@ static int adapter_up(struct btd_adapter *adapter, int dd)
 	}
 
 proceed:
-	if (dev_down == FALSE)
-		hci_send_cmd(dd, OGF_HOST_CTL, OCF_WRITE_SCAN_ENABLE,
-				1, &scan_mode);
+	if (scan_mode == SCAN_PAGE)
+		err = adapter_ops->set_connectable(adapter->dev_id);
+	else
+		err = adapter_ops->set_discoverable(adapter->dev_id);
+
+	if (err < 0)
+		return err;
 
 	if (adapter->initialized == FALSE) {
 		load_drivers(adapter);
@@ -2066,17 +2034,19 @@ proceed:
 
 		/* retrieve the active connections: address the scenario where
 		 * the are active connections before the daemon've started */
-		load_connections(adapter, dd);
+		load_connections(adapter);
 
 		adapter->initialized = TRUE;
 
-		adapter_update(adapter, 0, FALSE);
-
 		manager_add_adapter(adapter->path);
+
 	}
 
+	if (adapter->svc_cache)
+		adapter_update(adapter, 0);
+
 	if (dev_down) {
-		ioctl(dd, HCIDEVDOWN, adapter->dev_id);
+		adapter_ops->stop(adapter->dev_id);
 		adapter->off_requested = TRUE;
 		return 1;
 	} else
@@ -2094,7 +2064,6 @@ int adapter_start(struct btd_adapter *adapter)
 	struct hci_version ver;
 	uint8_t features[8];
 	int dd, err;
-	char name[249];
 
 	if (hci_devinfo(adapter->dev_id, &di) < 0)
 		return -errno;
@@ -2156,16 +2125,7 @@ int adapter_start(struct btd_adapter *adapter)
 		return err;
 	}
 
-	if (hci_read_local_name(dd, sizeof(name), name,
-						HCI_REQ_TIMEOUT) < 0) {
-		err = -errno;
-		error("Can't read local name on %s: %s (%d)",
-					adapter->path, strerror(errno), errno);
-		hci_close_dev(dd);
-		return err;
-	}
-
-	memcpy(dev->name, name, 248);
+	adapter_ops->read_name(adapter->dev_id);
 
 	if (!(features[6] & LMP_SIMPLE_PAIR))
 		goto setup;
@@ -2185,21 +2145,16 @@ int adapter_start(struct btd_adapter *adapter)
 setup:
 	hci_send_cmd(dd, OGF_LINK_POLICY, OCF_READ_DEFAULT_LINK_POLICY,
 								0, NULL);
+	hci_close_dev(dd);
 
-	if (hci_test_bit(HCI_INQUIRY, &di.flags)) {
-		debug("inquiry_cancel at adapter startup");
-		inquiry_cancel(dd, HCI_REQ_TIMEOUT);
-	} else if (!adapter->initialized && adapter->already_up) {
-		debug("periodic_inquiry_exit at adapter startup");
-		periodic_inquiry_exit(dd, HCI_REQ_TIMEOUT);
+	adapter_setup(adapter);
+
+	if (!adapter->initialized && adapter->already_up) {
+		debug("Stopping Inquiry at adapter startup");
+		adapter_ops->stop_discovery(adapter->dev_id);
 	}
 
-	adapter->state &= ~STD_INQUIRY;
-
-	adapter_setup(adapter, dd);
-	err = adapter_up(adapter, dd);
-
-	hci_close_dev(dd);
+	err = adapter_up(adapter);
 
 	info("Adapter %s has been enabled", adapter->path);
 
@@ -2222,16 +2177,10 @@ static void reply_pending_requests(struct btd_adapter *adapter)
 						HCI_OE_USER_ENDED_CONNECTION);
 	}
 
-	if (adapter->state & STD_INQUIRY) {
+	if (adapter->state & STD_INQUIRY || adapter->state & PERIODIC_INQUIRY) {
 		/* Cancel inquiry initiated by D-Bus client */
 		if (adapter->disc_sessions)
-			cancel_discovery(adapter);
-	}
-
-	if (adapter->state & PERIODIC_INQUIRY) {
-		/* Stop periodic inquiry initiated by D-Bus client */
-		if (adapter->disc_sessions)
-			cancel_periodic_discovery(adapter);
+			adapter_ops->stop_discovery(adapter->dev_id);
 	}
 }
 
@@ -2267,15 +2216,9 @@ int adapter_stop(struct btd_adapter *adapter)
 		adapter->disc_sessions = NULL;
 	}
 
-	if (adapter->found_devices) {
-		g_slist_foreach(adapter->found_devices,
-				(GFunc) dev_info_free, NULL);
-		g_slist_free(adapter->found_devices);
-		adapter->found_devices = NULL;
-	}
+	clear_found_devices_list(adapter);
 
 	if (adapter->oor_devices) {
-		g_slist_foreach(adapter->oor_devices, (GFunc) free, NULL);
 		g_slist_free(adapter->oor_devices);
 		adapter->oor_devices = NULL;
 	}
@@ -2313,18 +2256,17 @@ int adapter_stop(struct btd_adapter *adapter)
 	return 0;
 }
 
-int adapter_update(struct btd_adapter *adapter, uint8_t new_svc,
-							gboolean starting)
+int adapter_update(struct btd_adapter *adapter, uint8_t new_svc)
 {
 	struct hci_dev *dev = &adapter->dev;
-	int dd;
 	uint8_t svclass;
 
 	if (dev->ignore)
 		return 0;
 
-	if (starting || !adapter->initialized) {
-		adapter->svc_cache = new_svc;
+	if (!adapter->up) {
+		if (new_svc)
+			adapter->svc_cache = new_svc;
 		return 0;
 	}
 
@@ -2333,20 +2275,10 @@ int adapter_update(struct btd_adapter *adapter, uint8_t new_svc,
 	else
 		svclass = adapter->svc_cache;
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0) {
-		int err = -errno;
-		error("Can't open adapter %s: %s (%d)",
-					adapter->path, strerror(errno), errno);
-		return err;
-	}
-
 	if (svclass)
-		set_service_classes(dd, adapter->dev.class, svclass);
+		set_service_classes(adapter, svclass);
 
-	update_ext_inquiry_response(dd, dev);
-
-	hci_close_dev(dd);
+	update_ext_inquiry_response(adapter);
 
 	return 0;
 }
@@ -2383,15 +2315,13 @@ int adapter_set_class(struct btd_adapter *adapter, uint8_t *cls)
 	return 0;
 }
 
-int adapter_update_ssp_mode(struct btd_adapter *adapter, int dd, uint8_t mode)
+int adapter_update_ssp_mode(struct btd_adapter *adapter, uint8_t mode)
 {
 	struct hci_dev *dev = &adapter->dev;
 
 	dev->ssp_mode = mode;
 
-	update_ext_inquiry_response(dd, dev);
-
-	hci_close_dev(dd);
+	update_ext_inquiry_response(adapter);
 
 	return 0;
 }
@@ -2403,10 +2333,37 @@ static void adapter_free(gpointer user_data)
 	agent_destroy(adapter->agent, FALSE);
 	adapter->agent = NULL;
 
+	debug("adapter_free(%p)", adapter);
+
 	g_free(adapter->path);
 	g_free(adapter);
+}
 
-	return;
+struct btd_adapter *btd_adapter_ref(struct btd_adapter *adapter)
+{
+	adapter->ref++;
+
+	debug("btd_adapter_ref(%p): ref=%d", adapter, adapter->ref);
+
+	return adapter;
+}
+
+void btd_adapter_unref(struct btd_adapter *adapter)
+{
+	gchar *path;
+
+	adapter->ref--;
+
+	debug("btd_adapter_unref(%p): ref=%d", adapter, adapter->ref);
+
+	if (adapter->ref > 0)
+		return;
+
+	path = g_strdup(adapter->path);
+
+	g_dbus_unregister_interface(connection, path, ADAPTER_INTERFACE);
+
+	g_free(path);
 }
 
 struct btd_adapter *adapter_create(DBusConnection *conn, int id,
@@ -2428,7 +2385,8 @@ struct btd_adapter *adapter_create(DBusConnection *conn, int id,
 	}
 
 	adapter->dev_id = id;
-	adapter->state |= RESOLVE_NAME;
+	if (main_opts.name_resolv)
+		adapter->state |= RESOLVE_NAME;
 	adapter->path = g_strdup(path);
 	adapter->already_up = devup;
 
@@ -2440,15 +2398,14 @@ struct btd_adapter *adapter_create(DBusConnection *conn, int id,
 		return NULL;
 	}
 
-	return adapter;
+	return btd_adapter_ref(adapter);
 }
 
 void adapter_remove(struct btd_adapter *adapter)
 {
 	GSList *l;
-	char *path = g_strdup(adapter->path);
 
-	debug("Removing adapter %s", path);
+	debug("Removing adapter %s", adapter->path);
 
 	for (l = adapter->devices; l; l = l->next)
 		device_remove(l->data, connection, FALSE);
@@ -2457,18 +2414,10 @@ void adapter_remove(struct btd_adapter *adapter)
 	unload_drivers(adapter);
 
 	/* Return adapter to down state if it was not up on init */
-	if (adapter->up && !adapter->already_up) {
-		int dd = hci_open_dev(adapter->dev_id);
-		if (dd < 0)
-			goto done;
+	if (adapter->up && !adapter->already_up)
+		adapter_ops->stop(adapter->dev_id);
 
-		ioctl(dd, HCIDEVDOWN, adapter->dev_id);
-		hci_close_dev(dd);
-	}
-
-done:
-	g_dbus_unregister_interface(connection, path, ADAPTER_INTERFACE);
-	g_free(path);
+	btd_adapter_unref(adapter);
 }
 
 uint16_t adapter_get_dev_id(struct btd_adapter *adapter)
@@ -2499,9 +2448,10 @@ void adapter_set_state(struct btd_adapter *adapter, int state)
 
 	if (state & PERIODIC_INQUIRY || state & STD_INQUIRY)
 		discov_active = TRUE;
-	else if (adapter->disc_sessions && main_opts.inqmode)
-		adapter->scheduler_id = g_timeout_add_seconds(main_opts.inqmode,
-						(GSourceFunc) start_inquiry,
+	else if (adapter->disc_sessions && main_opts.discov_interval)
+		adapter->scheduler_id = g_timeout_add_seconds(
+						main_opts.discov_interval,
+						(GSourceFunc) adapter_start_inquiry,
 						adapter);
 
 	/* Send out of range */
@@ -2654,6 +2604,11 @@ void adapter_update_found_devices(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 	if (dev) {
 		if (rssi == dev->rssi)
 			return;
+
+		/* Out of range list update */
+		adapter->oor_devices = g_slist_remove(adapter->oor_devices,
+							dev);
+
 		goto done;
 	}
 
@@ -2698,53 +2653,27 @@ int adapter_remove_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr)
 void adapter_update_oor_devices(struct btd_adapter *adapter)
 {
 	GSList *l;
-	struct remote_dev_info *dev;
-	bdaddr_t tmp;
 
 	for (l = adapter->oor_devices; l; l = l->next) {
-		char *address = l->data;
-		struct remote_dev_info match;
+		char address[18];
+		const char *paddr = address;
+		struct remote_dev_info *dev = l->data;
 
-		memset(&match, 0, sizeof(struct remote_dev_info));
-		str2ba(address, &match.bdaddr);
+		ba2str(&dev->bdaddr, address);
 
 		g_dbus_emit_signal(connection, adapter->path,
 				ADAPTER_INTERFACE, "DeviceDisappeared",
-				DBUS_TYPE_STRING, &address,
+				DBUS_TYPE_STRING, &paddr,
 				DBUS_TYPE_INVALID);
 
-		dev = adapter_search_found_devices(adapter, &match);
-		if (!dev)
-			continue;
-
 		adapter->found_devices = g_slist_remove(adapter->found_devices, dev);
 		dev_info_free(dev);
 	}
 
-	g_slist_foreach(adapter->oor_devices, (GFunc) free, NULL);
 	g_slist_free(adapter->oor_devices);
 	adapter->oor_devices = NULL;
 
-	for (l = adapter->found_devices; l; l = l->next) {
-		dev = l->data;
-		baswap(&tmp, &dev->bdaddr);
-		adapter->oor_devices = g_slist_append(adapter->oor_devices,
-							batostr(&tmp));
-	}
-}
-
-void adapter_remove_oor_device(struct btd_adapter *adapter, char *peer_addr)
-{
-	GSList *l;
-
-	l = g_slist_find_custom(adapter->oor_devices, peer_addr,
-				(GCompareFunc) strcmp);
-	if (l) {
-		char *dev = l->data;
-		adapter->oor_devices = g_slist_remove(adapter->oor_devices,
-								dev);
-		g_free(dev);
-	}
+	adapter->oor_devices = g_slist_copy(adapter->found_devices);
 }
 
 void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
@@ -2752,7 +2681,6 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 	const gchar *path = adapter_get_path(adapter);
 	gboolean discoverable, pairable;
 	uint8_t real_class[3];
-	int dd;
 
 	if (adapter->scan_mode == scan_mode)
 		return;
@@ -2797,23 +2725,17 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 					ADAPTER_INTERFACE, "Pairable",
 					DBUS_TYPE_BOOLEAN, &pairable);
 
-	dd = hci_open_dev(adapter->dev_id);
-	if (dd < 0) {
-		error("HCI device open failed: hci%d", adapter->dev_id);
-		goto done;
-	}
-
 	memcpy(real_class, adapter->dev.class, 3);
 	if (adapter->svc_cache)
 		real_class[2] = adapter->svc_cache;
 
 	if (discoverable && adapter->pairable)
-		set_limited_discoverable(dd, real_class, TRUE);
+		adapter_ops->set_limited_discoverable(adapter->dev_id,
+							real_class, TRUE);
 	else if (!discoverable)
-		set_limited_discoverable(dd, real_class, FALSE);
+		adapter_ops->set_limited_discoverable(adapter->dev_id,
+							real_class, FALSE);
 
-	hci_close_dev(dd);
-done:
 	emit_property_changed(connection, path,
 				ADAPTER_INTERFACE, "Discoverable",
 				DBUS_TYPE_BOOLEAN, &discoverable);
@@ -3093,12 +3015,12 @@ int btd_register_adapter_ops(struct btd_adapter_ops *btd_adapter_ops)
 	return 0;
 }
 
-void btd_adapter_cleanup_ops()
+void btd_adapter_cleanup_ops(struct btd_adapter_ops *btd_adapter_ops)
 {
 	adapter_ops->cleanup();
 }
 
-int adapter_ops_setup()
+int adapter_ops_setup(void)
 {
 	if (!adapter_ops)
 		return -EINVAL;
